//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	Copyright (C) 1999-2000 Jnis Legzdi
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**	
//**************************************************************************

float air_control = 1.0 / 256.0;

#define INVERSECOLORMAP		32

//
//	PALETTE MACROS
//

// Palette indices.
// For damage/bonus red-/gold-shifts
#define STARTREDPALS		1
#define STARTBONUSPALS		9
#define NUMREDPALS			8
#define NUMBONUSPALS		4
// Radiation suit, green shift.
#define RADIATIONPAL		13

void MovePsprites(player_t* player);
void SetPlayerRunState(player_t* player);
void ChangeWeapon(player_t* player, int newweapon);
void DamageMobj(mobj_t* target,mobj_t* inflictor,mobj_t* source,int damage);
boolean G_CheckSpot(int playernum, mthing_t* mthing);
void P_SpawnPlayer(mthing_t* mthing);
void P_UseLines(player_t* player);

boolean		onground;

//==========================================================================
//
// 	ThrustPlayer
//
// 	Moves the given origin along a given angle.
//
//==========================================================================

void ThrustPlayer(player_t* player, angle_t angle, float move)
{
	player->mo->velocity.x += move * cos(angle) * frametime;
	player->mo->velocity.y += move * sin(angle) * frametime;
}

//==========================================================================
//
// 	CalcHeight
//
// 	Calculate the walking / running height adjustment
//
//==========================================================================

void CalcHeight(player_t* player)
{
    int			angle;
	float		bob;
    
    // Regular movement bobbing
    // (needs to be calculated for gun swing
    // even if not on ground)
    // OPTIMIZE: tablify angle
    // Note: a LUT allows for effects
    //  like a ramp with low health.
   	player->bob = (player->mo->velocity.x * player->mo->velocity.x
   				+  player->mo->velocity.y * player->mo->velocity.y)
   				* frametime * frametime;

   	player->bob /= 4.0;
	if (player->bob > MAXBOB)
		player->bob = MAXBOB;

    angle = ANG180 / 10 * level->tictime;
    bob = player->bob / 2.0 * sin(angle);

    
    // move viewheight
    if (player->playerstate == PST_LIVE)
    {
		player->viewheight += player->deltaviewheight;

		if (player->viewheight > VIEWHEIGHT)
		{
		    player->viewheight = VIEWHEIGHT;
		    player->deltaviewheight = 0.0;
		}

		if (player->viewheight < VIEWHEIGHT / 2.0)
		{
		    player->viewheight = VIEWHEIGHT / 2.0;
		    if (player->deltaviewheight <= 0.0)
				player->deltaviewheight = 0.00001;
		}
	
		if (player->deltaviewheight)
		{
		    player->deltaviewheight += 0.25;
		    if (!player->deltaviewheight)
				player->deltaviewheight = 0.00001;
		}
    }
	player->vieworg.z = player->mo->origin.z + player->viewheight + bob;

    if (player->vieworg.z > player->mo->ceilingz - 4.0)
		player->vieworg.z = player->mo->ceilingz - 4.0;
}

//==========================================================================
//
//  MovePlayer
//
//==========================================================================

void MovePlayer(player_t* player)
{
	float		forward;
	float		side;

    // Do not let the player control movement
    //  if not onground.
	onground = player->mo->origin.z <= player->mo->floorz ||
				(player->mo->flags2 & MF2_ONMOBJ);

	forward = player->forwardmove * 5.0;
	side = player->sidemove * 5.0;

	if (!onground)
	{
		forward *= air_control;
		side *= air_control;
	}

	if (forward)
	{
		ThrustPlayer(player, player->mo->angles.yaw, forward);
	}
	if (side)
	{
		ThrustPlayer(player, player->mo->angles.yaw - ANG90, side);
	}

    if (forward || side)
    {
		SetPlayerRunState(player);
	}

	if ((player->buttons & BT_JUMP) && onground && !player->jumpTics)
	{
       	//	Jump
		player->mo->velocity.z = 9.0 * 35.0;
		player->jumpTics = 18;
	}
}

//==========================================================================
//
//	CheckWaterJump
//
//==========================================================================

void CheckWaterJump(player_t *player)
{
	TVec	start;
	TVec	end;
	TVec	vforward;

	// check for a jump-out-of-water
	AngleVector(&player->mo->angles, &vforward);
	start = player->mo->origin;
	start.z += player->mo->height * 0.5 + 8.0;
	vforward.z = 0.0;
	vforward = Normalize(vforward);
	end = start + vforward * 24.0;
	if (!TraceLine(start, end))
	{
		// solid at waist
		start.z = player->mo->origin.z + player->mo->height;
		end = start + vforward * 24.0;
		player->movedir = trace_plane_normal * -50.0;
		if (TraceLine(start, end))
		{
			// open at eye level
			player->mo->flags |= MF_WATERJUMP;
			player->mo->velocity.z = 320.0;
			player->mo->reactiontime = 70;	// safety net
		}
	}
}

//==========================================================================
//
//  WaterMove
//
//==========================================================================

void WaterMove(player_t* player)
{
	float		forward;
	float		side;
	TVec		vforward;
	TVec		vright;
	TVec		vup;
	TVec		wishvel;

    // Do not let the player control movement
    //  if not onground.
	onground = player->mo->origin.z <= player->mo->floorz ||
				(player->mo->flags2 & MF2_ONMOBJ);

	AngleVectors(&player->mo->angles, &vforward, &vright, &vup);

	forward = player->forwardmove;
	side = player->sidemove;

	wishvel = forward * vforward + side * vright;
	if (!forward && !side/* && !cmd.upmove*/)
		wishvel.z -= 60.0;		// drift towards bottom
//	else
//		wishvel.z += cmd.upmove;

	player->mo->velocity += 3.5 * frametime * wishvel;

    if (forward || side)
    {
		SetPlayerRunState(player);
	}

	if (player->buttons & BT_JUMP)
	{
		if (player->mo->watertype == CONTENTS_WATER)
			player->mo->velocity.z = 100.0;
		else if (player->mo->watertype == CONTENTS_NUKAGE ||
			player->mo->watertype == CONTENTS_SLIME)
			player->mo->velocity.z = 80.0;
		else
			player->mo->velocity.z = 50.0;
	}
	CheckWaterJump(player);
}

//==========================================================================
//
//	WaterJump
//
//==========================================================================

void WaterJump(player_t *player)
{
	if (!player->mo->reactiontime || !player->mo->waterlevel)
	{
		player->mo->flags &= ~MF_WATERJUMP;
		player->mo->reactiontime = 0;
	}
	player->mo->velocity.x = player->movedir.x;
	player->mo->velocity.y = player->movedir.y;
}

//==========================================================================
//
//  PlayerInSpecialSector
//
//	Called every tic frame that the player origin is in a special sector
//
//==========================================================================

void PlayerInSpecialSector(player_t* player)
{
    sector_t*	sector;
	
    sector = player->mo->subsector->sector;

    // Falling, not all the way down yet?
    if (player->mo->origin.z != sector->floorheight)
		return;

	if (sector->special & SECRET_MASK)
	{
		// SECRET SECTOR
		player->secretcount++;
		centerprint(player, "You found a secret area");
		sector->special &= ~SECRET_MASK;
	}

    // Has hitten ground.
	switch (sector->special)
	{
      case 64+5:
		// HELLSLIME DAMAGE
		if (!player->powers[pw_ironfeet])
		    if (!(level->tictime & 0x1f))
				DamageMobj(player->mo, NULL, NULL, 10);
		break;
	
      case 64+7:
		// NUKAGE DAMAGE
		if (!player->powers[pw_ironfeet])
		    if (!(level->tictime & 0x1f))
				DamageMobj(player->mo, NULL, NULL, 5);
		break;
	
      case 64+16:
		// SUPER HELLSLIME DAMAGE
      case 64+4:
		// STROBE HURT
		if (!player->powers[pw_ironfeet] || (P_Random()<5))
		{
		    if (!(level->tictime & 0x1f))
				DamageMobj(player->mo, NULL, NULL, 20);
		}
		break;
			
      case 64+11:
		// EXIT SUPER DAMAGE! (for E1M8 finale)
		player->cheats &= ~CF_GODMODE;

		if (!(level->tictime & 0x1f))
		    DamageMobj(player->mo, NULL, NULL, 20);

		if (player->health <= 10)
		    G_ExitLevel();
		break;
			
      default:
		break;
     };
}

//==========================================================================
//
//	PlayerInContents
//
//==========================================================================

void PlayerInContents(player_t *player)
{
	if (!player->mo->waterlevel)
	{
		return;
	}
	switch (player->mo->watertype)
	{
	 case CONTENTS_NUKAGE:
		// NUKAGE DAMAGE
		if (!player->powers[pw_ironfeet])
			if (!(level->tictime & 0x1f))
				DamageMobj(player->mo, NULL, NULL, 5);
		break;

	 case CONTENTS_SLIME:
		// HELLSLIME DAMAGE
		if (!player->powers[pw_ironfeet])
			if (!(level->tictime & 0x1f))
				DamageMobj(player->mo, NULL, NULL, 10);
		break;

	 case CONTENTS_LAVA:
		// SUPER HELLSLIME DAMAGE
		if (!player->powers[pw_ironfeet] || (P_Random() < 5))
		{
			if (!(level->tictime & 0x1f))
				DamageMobj(player->mo, NULL, NULL, 20);
		}
		break;
	}
}

//==========================================================================
//
// 	DeathPlayerThink
//
// 	Fall on your face when dying. Decrease POV height to floor height.
//
//==========================================================================

void DeathPlayerThink(player_t* player)
{
    angle_t		angle;
    angle_t		delta;

    MovePsprites(player);
	
	onground = player->mo->origin.z <= player->mo->floorz ||
				(player->mo->flags2 & MF2_ONMOBJ);

	if (player->mo->waterlevel > 1)
	{
		player->mo->velocity.z = -60.0;	// drift towards bottom
	}

    // fall to the ground
    if (player->viewheight > 6.0)
		player->viewheight -= 1.0;

    if (player->viewheight < 6.0)
		player->viewheight = 6.0;

    player->deltaviewheight = 0.0;
    CalcHeight(player);
	
    if (player->attacker && player->attacker != player->mo)
    {
		angle = atan2(player->attacker->origin.y - player->mo->origin.y,
					  player->attacker->origin.x - player->mo->origin.x);
	
		delta = angle - player->mo->angles.yaw;
	
		if (delta < ANG5 || delta > -ANG5)
		{
		    // Looking at killer,
	    	//  so fade damage flash down.
		    player->mo->angles.yaw = angle;

		    if (player->damagecount)
				player->damagecount--;
		}
		else if (delta < ANG180)
		    player->mo->angles.yaw += ANG5;
		else
		    player->mo->angles.yaw -= ANG5;
    }
    else if (player->damagecount)
		player->damagecount--;
	

    if (player->buttons & BT_USE)
		player->playerstate = PST_REBORN;
}

//==========================================================================
//
//	PlayerImpulse
//
//==========================================================================

void PlayerImpulse(player_t *player)
{
	int		impulse;

	impulse = player->impulse;
	player->impulse = 0;

    if (impulse <= 7)
    {
		ChangeWeapon(player, impulse - 1);
	}
    else if (impulse == 16)
    {
		//	Suicide
		DamageMobj(player->mo, NULL, NULL, 10000);
    }
}

//==========================================================================
//
// 	PlayerThink
//
//==========================================================================

void PlayerThink(player_t* player)
{
    // fixme: do this in the cheat code
    if (player->cheats & CF_NOCLIP)
		player->mo->flags |= MF_NOCLIP;
    else
		player->mo->flags &= ~MF_NOCLIP;
    
    if (player->mo->flags & MF_JUSTATTACKED)
    {
		player->forwardmove = 100.0;
		player->sidemove = 0.0;
		player->mo->flags &= ~MF_JUSTATTACKED;
    }

    if (player->playerstate == PST_DEAD)
    {
		DeathPlayerThink(player);
		return;
    }
    
	if (player->jumpTics)
	{
		player->jumpTics--;
	}

    // Move around.
    // Reactiontime is used to prevent movement
    //  for a bit after a teleport.
    if (player->mo->reactiontime)
	{
		player->mo->reactiontime--;
		if (player->mo->flags & MF_WATERJUMP)
			WaterJump(player);
	}
    else
	{
		if (player->mo->waterlevel > 1)
			WaterMove(player);
		else
			MovePlayer(player);
	}
    
    CalcHeight(player);

    if (player->mo->subsector->sector->special)
		PlayerInSpecialSector(player);
	PlayerInContents(player);

    // Check for weapon change.
    if (player->impulse)
    {
		PlayerImpulse(player);
	}

    // check for use
    if (player->buttons & BT_USE)
    {
		if (!player->usedown)
		{
		    P_UseLines(player);
		    player->usedown = true;
		}
    }
    else
	{
		player->usedown = false;
	}

    //	cycle psprites
    MovePsprites(player);
    
    //	Counters, time dependend power ups.

    //	Strength counts up to diminish fade.
    if (player->powers[pw_strength])
		player->powers[pw_strength]++;
		
    if (player->powers[pw_invulnerability])
		player->powers[pw_invulnerability]--;

    if (player->powers[pw_invisibility])
		if (! --player->powers[pw_invisibility])
		    player->mo->translucency = 0;
			
    if (player->powers[pw_infrared])
		player->powers[pw_infrared]--;
		
    if (player->powers[pw_ironfeet])
		player->powers[pw_ironfeet]--;

    if (player->damagecount)
		player->damagecount--;
		
    if (player->bonuscount)
		player->bonuscount--;

    // Handling colormaps.
    if (player->powers[pw_invulnerability])
    {
		if (player->powers[pw_invulnerability] > 4 * 32
			    || (player->powers[pw_invulnerability] & 8) )
		    player->fixedcolormap = INVERSECOLORMAP;
		else
		    player->fixedcolormap = 0;
    }
    else if (player->powers[pw_infrared])
    {
		if (player->powers[pw_infrared] > 4 * 32
		    || (player->powers[pw_infrared] & 8))
		{
		    // almost full bright
		    player->fixedcolormap = 1;
		}
		else
		    player->fixedcolormap = 0;
    }
    else
		player->fixedcolormap = 0;
}

//==========================================================================
//
//  SB_PaletteFlash
//
//	Do red-/gold-shifts from damage/items
//
//==========================================================================

void SB_PaletteFlash(player_t *player)
{
    int		cnt;
    int		bzc;

	player->cshifts[CSHIFT_CONTENTS] = 0;
	if (player->mo->waterlevel == 3)
	{
		switch (player->mo->watertype)
		{
		 case CONTENTS_WATER:
 			player->cshifts[CSHIFT_CONTENTS] = CSHIFT(130, 80, 50, 128);
			break;

		 case CONTENTS_BLOOD:
 			player->cshifts[CSHIFT_CONTENTS] = CSHIFT(160, 16, 16, 150);
			break;

		 case CONTENTS_NUKAGE:
 			player->cshifts[CSHIFT_CONTENTS] = CSHIFT(50, 255, 50, 150);
			break;

		 case CONTENTS_SLIME:
 			player->cshifts[CSHIFT_CONTENTS] = CSHIFT(0, 25, 5, 150);
			break;

		 case CONTENTS_LAVA:
 			player->cshifts[CSHIFT_CONTENTS] = CSHIFT(255, 80, 0, 150);
			break;
		}
	}

    cnt = player->damagecount;

    if (player->powers[pw_strength])
    {
		// slowly fade the berzerk out
  		bzc = 12 - (player->powers[pw_strength] >> 6);

		if (bzc > cnt)
	    	cnt = bzc;
    }
	
    if (cnt)
    {
		player->palette = (cnt + 7) >> 3;
	
		if (player->palette >= NUMREDPALS)
	    	player->palette = NUMREDPALS - 1;

		player->palette += STARTREDPALS;
    }
    else if (player->bonuscount)
    {
		player->palette = (player->bonuscount + 7) >> 3;

		if (player->palette >= NUMBONUSPALS)
	    	player->palette = NUMBONUSPALS - 1;

		player->palette += STARTBONUSPALS;
    }
    else if (player->powers[pw_ironfeet] > 4 * 32
	      || player->powers[pw_ironfeet] & 8)
		player->palette = RADIATIONPAL;
    else
		player->palette = 0;
}

//==========================================================================
//
//	SetViewPos
//
//==========================================================================

void SetViewPos(player_t *player)
{
	player->vieworg.x = player->mo->origin.x;
	player->vieworg.y = player->mo->origin.y;
	SB_PaletteFlash(player);

    if (player->w_ammo == am_noammo)
		player->ready_ammo = 1994; // means "n/a"
    else
		player->ready_ammo = player->ammo[player->w_ammo];
}

//==========================================================================
//
//	G_PlayerReborn
//
//	Called after a player dies almost everything is cleared and initialized
//
//==========================================================================

void G_PlayerReborn(int player)
{
    player_t*	p; 
	 
    p = &players[player];

	//	Clear player struct
    ClearPlayer(p);

	//	Set initial data
    p->usedown = p->attackdown = true;	// don't do anything immediately 
    p->playerstate = PST_LIVE;       
    p->health = INITIAL_HEALTH;
    p->readyweapon = p->pendingweapon = wp_pistol;
	set_weapon_pistol(p);
    p->weaponowned[wp_fist] = true; 
    p->weaponowned[wp_pistol] = true; 
    p->ammo[am_clip] = INITIAL_AMMO;
}

//==========================================================================
//
//  G_PlayerExitMap
//
//	Call when a player completes a level.
//
//==========================================================================

void G_PlayerExitMap(int player, boolean clusterChange)
{ 
    player_t*	p;
	int			i;
	 
    p = &players[player]; 

	if (clusterChange)
	{
	    p->keys = 0;
	}

	for (i = 0; i < NUMPOWERS; i++)
    {
	    p->powers[i] = 0;
	}
	p->mo->angles.pitch = 0;
    p->mo->translucency = 0;	// cancel invisibility
    p->extralight = 0;			// cancel gun flashes 
    p->fixedcolormap = 0;		// cancel ir gogles 
    p->damagecount = 0;			// no palette changes 
    p->bonuscount = 0; 
}

//==========================================================================
//
//	G_DeathMatchSpawnPlayer
//
//	Spawns a player at one of the random death match spots called at level
// load and each death
//
//==========================================================================

void G_DeathMatchSpawnPlayer(int playernum)
{ 
    int             i;
    int				j;
	 
	// This check has been moved to p_setup.c:P_LoadThings()
//	if (selections < 4)
//		I_Error("Only %i deathmatch spots, 4 required", selections);
 
    for (j=0 ; j<20 ; j++) 
    { 
		i = P_Random() % level->numdeathmatchstarts;
		if (G_CheckSpot(playernum, &level->deathmatchstarts[i]))
		{
	    	level->deathmatchstarts[i].type = playernum + 1;
	    	P_SpawnPlayer(&level->deathmatchstarts[i]);
	    	return;
		}
    } 
 
    // no good spot, so the player will probably get stuck 
    P_SpawnPlayer(&level->playerstarts[playernum]);
} 

//==========================================================================
//
//  NetGameReborn
//
//==========================================================================

void NetGameReborn(int playernum)
{
    int			i;

	// respawn at the start

	// first dissasociate the corpse
	players[playernum].mo->player = NULL;
		 
	// spawn at random spot if in death match
	if (deathmatch)
	{
    	G_DeathMatchSpawnPlayer(playernum);
    	return;
	}
		 
	if (G_CheckSpot(playernum, &level->playerstarts[RebornPosition * MAXPLAYERS + playernum]))
	{
    	P_SpawnPlayer(&level->playerstarts[RebornPosition * MAXPLAYERS + playernum]);
    	return;
	}
	
	// try to spawn at one of the other players spots
	for (i=0 ; i<MAXPLAYERS ; i++)
	{
    	if (G_CheckSpot(playernum, &level->playerstarts[RebornPosition * MAXPLAYERS + i]))
    	{
			level->playerstarts[RebornPosition * MAXPLAYERS + i].type = playernum + 1;// fake as other player
			P_SpawnPlayer(&level->playerstarts[RebornPosition * MAXPLAYERS + i]);
			level->playerstarts[RebornPosition * MAXPLAYERS + i].type = i + 1;		// restore
			return;
       	}
    	// he's going to be inside something.  Too bad.
	}
	P_SpawnPlayer(&level->playerstarts[RebornPosition * MAXPLAYERS + playernum]);
}

//==========================================================================
//
//	SpawnClient
//
//==========================================================================

void SpawnClient(int i)
{
   	if (deathmatch)
   	{
		G_DeathMatchSpawnPlayer(i);
	}
	else
	{
		P_SpawnPlayer(&level->playerstarts[RebornPosition * MAXPLAYERS + i]);
	}
}

//==========================================================================
//
//	DisconnectClient
//
//==========================================================================

void DisconnectClient(player_t *player)
{
	if (player->mo)
	{
		player->mo->player = NULL;
		DamageMobj(player->mo, NULL, NULL, 10000);
	}
   	bprint("%s left the game\n", ARR2STR(player->name));
	StartSoundName(NULL, "Chat");
}

//==========================================================================
//
//	UserinfoChanged
//
//==========================================================================

void UserinfoChanged(player_t *player)
{
	if (!player->spawned)
	{
		return;
	}
	player->mo->translation = player->color;
}


